/* This file has been generated with generate-header-names.py script, do not edit */

#pragma once

typedef enum {
        SOUP_HEADER_ACCEPT,
        SOUP_HEADER_ACCEPT_CHARSET,
        SOUP_HEADER_ACCEPT_ENCODING,
        SOUP_HEADER_ACCEPT_LANGUAGE,
        SOUP_HEADER_ACCEPT_RANGES,
        SOUP_HEADER_ACCESS_CONTROL_ALLOW_CREDENTIALS,
        SOUP_HEADER_ACCESS_CONTROL_ALLOW_HEADERS,
        SOUP_HEADER_ACCESS_CONTROL_ALLOW_METHODS,
        SOUP_HEADER_ACCESS_CONTROL_ALLOW_ORIGIN,
        SOUP_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS,
        SOUP_HEADER_ACCESS_CONTROL_MAX_AGE,
        SOUP_HEADER_ACCESS_CONTROL_REQUEST_HEADERS,
        SOUP_HEADER_ACCESS_CONTROL_REQUEST_METHOD,
        SOUP_HEADER_AGE,
        SOUP_HEADER_AUTHENTICATION_INFO,
        SOUP_HEADER_AUTHORIZATION,
        SOUP_HEADER_CACHE_CONTROL,
        SOUP_HEADER_CONNECTION,
        SOUP_HEADER_CONTENT_DISPOSITION,
        SOUP_HEADER_CONTENT_ENCODING,
        SOUP_HEADER_CONTENT_LANGUAGE,
        SOUP_HEADER_CONTENT_LENGTH,
        SOUP_HEADER_CONTENT_LOCATION,
        SOUP_HEADER_CONTENT_RANGE,
        SOUP_HEADER_CONTENT_SECURITY_POLICY,
        SOUP_HEADER_CONTENT_SECURITY_POLICY_REPORT_ONLY,
        SOUP_HEADER_CONTENT_TYPE,
        SOUP_HEADER_COOKIE,
        SOUP_HEADER_COOKIE2,
        SOUP_HEADER_CROSS_ORIGIN_RESOURCE_POLICY,
        SOUP_HEADER_DNT,
        SOUP_HEADER_DATE,
        SOUP_HEADER_DEFAULT_STYLE,
        SOUP_HEADER_ETAG,
        SOUP_HEADER_EXPECT,
        SOUP_HEADER_EXPIRES,
        SOUP_HEADER_HOST,
        SOUP_HEADER_IF_MATCH,
        SOUP_HEADER_IF_MODIFIED_SINCE,
        SOUP_HEADER_IF_NONE_MATCH,
        SOUP_HEADER_IF_RANGE,
        SOUP_HEADER_IF_UNMODIFIED_SINCE,
        SOUP_HEADER_KEEP_ALIVE,
        SOUP_HEADER_LAST_EVENT_ID,
        SOUP_HEADER_LAST_MODIFIED,
        SOUP_HEADER_LINK,
        SOUP_HEADER_LOCATION,
        SOUP_HEADER_ORIGIN,
        SOUP_HEADER_PING_FROM,
        SOUP_HEADER_PING_TO,
        SOUP_HEADER_PRAGMA,
        SOUP_HEADER_PROXY_AUTHENTICATE,
        SOUP_HEADER_PROXY_AUTHENTICATION_INFO,
        SOUP_HEADER_PROXY_AUTHORIZATION,
        SOUP_HEADER_PURPOSE,
        SOUP_HEADER_RANGE,
        SOUP_HEADER_REFERER,
        SOUP_HEADER_REFERRER_POLICY,
        SOUP_HEADER_REFRESH,
        SOUP_HEADER_SEC_WEBSOCKET_ACCEPT,
        SOUP_HEADER_SEC_WEBSOCKET_EXTENSIONS,
        SOUP_HEADER_SEC_WEBSOCKET_KEY,
        SOUP_HEADER_SEC_WEBSOCKET_PROTOCOL,
        SOUP_HEADER_SEC_WEBSOCKET_VERSION,
        SOUP_HEADER_SERVER,
        SOUP_HEADER_SERVER_TIMING,
        SOUP_HEADER_SERVICE_WORKER,
        SOUP_HEADER_SERVICE_WORKER_ALLOWED,
        SOUP_HEADER_SET_COOKIE,
        SOUP_HEADER_SET_COOKIE2,
        SOUP_HEADER_SOURCEMAP,
        SOUP_HEADER_TE,
        SOUP_HEADER_TIMING_ALLOW_ORIGIN,
        SOUP_HEADER_TRAILER,
        SOUP_HEADER_TRANSFER_ENCODING,
        SOUP_HEADER_UPGRADE,
        SOUP_HEADER_UPGRADE_INSECURE_REQUESTS,
        SOUP_HEADER_USER_AGENT,
        SOUP_HEADER_VARY,
        SOUP_HEADER_VIA,
        SOUP_HEADER_WWW_AUTHENTICATE,
        SOUP_HEADER_X_CONTENT_TYPE_OPTIONS,
        SOUP_HEADER_X_DNS_PREFETCH_CONTROL,
        SOUP_HEADER_X_FRAME_OPTIONS,
        SOUP_HEADER_X_SOURCEMAP,
        SOUP_HEADER_X_TEMP_TABLET,
        SOUP_HEADER_X_XSS_PROTECTION,

        SOUP_HEADER_UNKNOWN
} SoupHeaderName;

SoupHeaderName soup_header_name_from_string (const char    *str);
const char    *soup_header_name_to_string   (SoupHeaderName name);
